import json
import os

from jsonschema import validate, ValidationError

from mco_agent.exceptions import InvalidRPCData, ImproperlyConfigured


class ProtocolMessage:

    # Maps protocols to the correct ProtocolMessage subclass
    # Filled in by calls to register_protocol
    _protocols = {}

    _protocol = None
    _schema = None

    def __init__(
            self,
            **kwargs
    ):
        if self._schema is None:
            raise ImproperlyConfigured("Must call {0}.load_schema before constructing an object".format(
                self.__class__.__name__))

        self._properties = {}
        for prop in self._schema["properties"]:
            self._properties[prop] = kwargs.get(prop, None)

    def __getattr__(self, item):
        if item not in self._properties:
            raise AttributeError(item)

        return self._properties[item]

    @classmethod
    def load_schema(cls):
        if cls._schema is not None:
            return

        schema_dir = os.path.join(os.path.dirname(__file__), 'schemas')

        with open(os.path.join(schema_dir, '{0}.json'.format(cls._protocol)), 'r') as fp:
            cls._schema = json.load(fp)

    @classmethod
    def get_protocol(cls, protocol_name):
        if protocol_name not in cls._protocols:
            raise InvalidRPCData("Unsupported message protocol {0}".format(protocol_name))

        protocol = cls._protocols[protocol_name]
        return protocol

    @classmethod
    def from_json(cls, request):
        message = json.loads(request)
        return cls.from_dict(message)

    @classmethod
    def from_dict(cls, message):
        """ Constructs a ProtocolMessage object from a dictionary of fields

        :param message: dict Message fields received
        """
        cls.load_schema()

        try:
            validate(message, cls._schema)
        except ValidationError as e:
            # Use of jsonschema is an implementation detail, so convert this error
            # into our own exception type
            raise InvalidRPCData(str(e))

        # Clone the message so as not to modify the original
        fields = message.copy()

        # Allow subclasses to override the fields as required
        cls.parse_message_hook(fields)

        obj = cls(**fields)
        return obj

    @classmethod
    def parse_message_hook(cls, fields):
        """ Hook to override how the message is converted into a protocol object

        Subclasses may override this method in order to customise the arguments to the constructor,
        for example converting a nested object into another ProtocolMessage object

        :param fields: dict Populated list of fields to pass to the ProtocolMessage constructor
        :return: None
        """
        pass

    @staticmethod
    def create_reply():
        """ Returns a reply object appropriate for this protocol

        :return: Reply
        """
        raise InvalidRPCData('Method should only be called for a ProtocolMessage subclass')

    @classmethod
    def register_protocol(cls):
        """ Registers the decorated class with the given protocol name so the correct
            ProtocolMessage object can be constructed when a message is received

        :return:
        """

        # noinspection PyProtectedMember
        def decorator(protocol_cls):
            if protocol_cls._protocol is None:
                raise ImproperlyConfigured('ProtocolMessage classes must define _protocol name')

            cls._protocols[protocol_cls._protocol] = protocol_cls
            return protocol_cls

        return decorator


@ProtocolMessage.register_protocol()
class ExternalActivationRequest(ProtocolMessage):

    _protocol = 'io.choria.mcorpc.external.v1.activation_request'

    @staticmethod
    def create_reply():
        return ActivationReply()


@ProtocolMessage.register_protocol()
class ExternalRPCRequest(ProtocolMessage):

    _protocol = 'io.choria.mcorpc.external.v1.rpc_request'

    @staticmethod
    def create_reply():
        return ActionReply()


class Reply:

    def __init__(self, *args, **kwargs):
        pass

    def successful(self):
        return True


class ActionReply(Reply):

    _protocol = 'io.choria.mcorpc.external.v1.rpc_reply'

    def __init__(self, statuscode=0, statusmsg='', data=None, disableresponse=False, *args, **kwargs):
        Reply.__init__(self, *args, **kwargs)
        self.statuscode = statuscode
        self.statusmsg = statusmsg
        self.data = data
        self.disableresponse = disableresponse

        if not self.data:
            self.data = {}

    def to_json(self):
        message = json.dumps({
            'statuscode': self.statuscode,
            'statusmsg': self.statusmsg,
            'data': self.data,
        })
        return message

    def fail(self, code, message):
        self.statuscode = code
        self.statusmsg = message

    def successful(self):
        return self.statuscode == 0


class ActivationReply(Reply):

    _protocol = 'io.choria.mcorpc.external.v1.activation_reply'

    def __init__(self, *args, **kwargs):
        Reply.__init__(self, *args, **kwargs)
        self.activate = True

    def fail(self):
        self.activate = False

    def to_json(self):
        message = json.dumps({
            'activate': self.activate,
        })
        return message



import logging

from mco_agent.config import AgentConfig
from mco_agent.exceptions import UnknownRPCAction


class Agent:

    _actions = {}

    def __init__(self, request, reply):
        self.agent_name = self.__class__.__name__.lower()
        self.request = request
        self.reply = reply
        self.config = None

        self.logger = logging.getLogger('mcorpc.{0}'.format(self.agent_name))

    def load_config(self):
        self.config = AgentConfig(self.agent_name)
        self.config.read_config()

    # noinspection PyUnusedLocal
    @staticmethod
    def should_activate():
        """ Indicates whether the agent should be functional on this host

        Defaults to True which means the agent is unconditionally active. Agent subclasses
        can implement whatever tests are appropriate and return False if the agent should
        not be activated.

        :param self:
        :return: bool
        """
        return True

    def run(self):
        action_name = self.request.action
        if action_name not in self._actions:
            raise UnknownRPCAction(action_name)

        action_method = self._actions[action_name]
        action_method(self)


def register_actions(cls):
    """ Registers all marked methods in the agent class

    :param cls: Agent Subclass of Agent containing methods decorated with @action
    """
    for name, method in cls.__dict__.items():
        if hasattr(method, "_register_action"):
            cls._actions[name] = method
    return cls


def action(method):
    """ Marks an agent instance method to be registered as an action

    :param method:
    :return:
    """

    method._register_action = True

    return method



from mco_agent.agent import Agent, action, register_actions
from mco_agent.dispatcher import dispatch
from mco_agent.version import __version__

__all__ = ['__version__', 'Agent', 'action', 'dispatch', 'register_actions']



class AgentException(Exception):
    """ Base exception"""
    description = 'Unknown error'
    statuscode = 5

    def __str__(self):
        return '{0}: {1}'.format(self.description, ' '.join(self.args))


class RPCAborted(AgentException):
    description = "RPC Aborted"
    statuscode = 1


class UnknownRPCAction(AgentException):
    description = "Unknown RPC Action"
    statuscode = 2


class InvalidRPCData(AgentException):
    description = "Invalid Data"
    statuscode = 4


class ImproperlyConfigured(AgentException):
    description = "RPC Aborted"


class InactiveAgent(AgentException):
    description = "Agent is not activated"



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version_info = {}
with open(os.path.join(os.path.dirname(__file__), "mco_agent", "version.py")) as fp:
    exec(fp.read(), version_info)
pkg_version = version_info['__version__']

setuptools.setup(
    name="py-mco-agent",
    version=pkg_version,
    author="Ben Roberts",
    author_email="me@benroberts.net",
    description="A library for implementing Mcollecting agents in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/optiz0r/py-mco-agent",
    packages=setuptools.find_packages(exclude=['tests*']),
    package_data = {
        'mco_agent': ['schemas/*.json'],
    },
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.6',
)



import os


class AgentConfig:
    """ Class which handles reading settings from the agent configuration file
    """

    CONFIG_DIR = '/etc/puppetlabs/mcollective/plugin.d/'

    def __init__(self, agent_name):
        self.agent_name = agent_name
        self.config_file = os.path.join(os.environ.get('CHORIA_EXTERNAL_CONFIG', self.CONFIG_DIR), '{0}.cfg'.format(self.agent_name))
        self.config = {}

    def config_exists(self):
        return os.path.exists(self.config_file)

    def read_config(self):
        """ Reads the configuration in from disk
        """
        if not self.config_exists():
            return

        with open(self.config_file, 'r') as fp:
            contents = fp.readlines()
            self.config = self.parse_config(contents)

    @staticmethod
    def parse_config(contents):
        """ Parses the configuration data

        :param contents: list[str] List of lines from the configuration file
        :return: dict[str,str]
        """
        config = {}

        for line in contents:
            line = line.lstrip().rstrip('\r\n')

            # Ignore comments
            if line.startswith('#'):
                continue

            # Ignore lines which do not contain key=value pairs
            if '=' not in line:
                continue

            key, _, value = line.partition('=')

            config[key.strip()] = value.lstrip()

        return config

    def __contains__(self, key):
        return key in self.config

    def __getitem__(self, key):
        return self.config[key]

    def get(self, key, default_value=None):
        return self.config.get(key, default_value)



from __future__ import print_function

import logging
import os
import sys

from mco_agent.agent import Agent
from mco_agent.exceptions import ImproperlyConfigured, InactiveAgent, RPCAborted, AgentException
from mco_agent.protocol import ExternalRPCRequest, ExternalActivationRequest, ProtocolMessage


def dispatch(agent_cls):
    """ Processes an agent, parsing a request from stdin and marshalling the response back to stdout

    If agent is specified, bypasses plugin discovery and directly runs the action on the agent

    :param agent_cls: Agent Subclass of Agent which implements actions
    :return:
    """
    setup_logger(enable_debug=os.environ.get('CHORIA_EXTERNAL_DEBUG', '0'))

    protocol_name = os.environ.get('CHORIA_EXTERNAL_PROTOCOL', None)
    if not protocol_name:
        print("Unknown protocol", file=sys.stderr)
        exit(1)

    protocol = ProtocolMessage.get_protocol(protocol_name)
    reply = protocol.create_reply()

    try:
        if not issubclass(agent_cls, Agent):
            raise ImproperlyConfigured("Object {0} is not an Agent subclass".format(agent_cls.__class__.__name__))

        request = read_request(protocol, os.environ.get('CHORIA_EXTERNAL_REQUEST', None))

        agent = agent_cls(
            request=request,
            reply=reply
        )
        agent.load_config()

        if isinstance(request, ExternalActivationRequest):
            reply.activate = agent.should_activate() is True

        elif isinstance(request, ExternalRPCRequest):
            if not agent.should_activate():
                raise InactiveAgent()

            try:
                agent.run()
            except Exception as e:
                raise RPCAborted(str(e))

    except AgentException as e:
        reply.fail(e.statuscode, '{0}: {1}'.format(e.description, str(e)))

    write_reply(os.environ.get('CHORIA_EXTERNAL_REPLY', None), reply)

    if not reply.successful:
        exit(1)


def setup_logger(enable_debug=False):
    """ Configures logging for the agent

    By default, info logging is sent to stdout (displayed on request),
    warning and errors are sent to stdout (displayed always).

    To minimise output from other libraries, logging is disabled for all but
    the mcorpc hierarchy.

    Agents should log to 'mcorpc.agentname'
    """
    class InfoFilter(logging.Filter):
        """ Filters out events more serious than INFO so that only low-severity messages are recordsd
        """
        def filter(self, rec):
            return rec.levelno <= logging.INFO

    # Disable all logging from other libraries by default
    root_logger = logging.getLogger('.')
    root_logger.setLevel(100)

    logger = logging.getLogger('mcorpc')
    logger.setLevel(logging.DEBUG)

    info = logging.StreamHandler(sys.stdout)
    info.addFilter(InfoFilter())
    if enable_debug != '0':
        info.setLevel(logging.DEBUG)
    else:
        info.setLevel(logging.INFO)

    warn = logging.StreamHandler(sys.stderr)
    warn.setLevel(logging.WARNING)

    root_logger.addHandler(info)
    root_logger.addHandler(warn)


def read_request(protocol, request_file):
    """ Reads in the request

    Supports:
    - Reading request from a file specified by command line argument
    - Reading request from a file specified by environment variable
    - Reading request from stdin

    :param protocol: ProtocolMessage
    :param request_file:
    :return:
    """
    if request_file:
        with open(request_file, 'r') as fp:
            request_data = fp.read()
    else:
        request_data = sys.stdin.read()

    request = protocol.from_json(request_data)

    return request


def write_reply(reply_file, reply):
    if reply_file:
        with open(reply_file, 'w') as fp:
            fp.write(reply.to_json())
    else:
        print(reply.to_json())
